
/* rtws device, read the tws on the back of the case. 

GPIO17- wire 1 bit 0 inverted
GPIO27- wire 2 bit 1 inverted
GPIO22- wire 4 bit 2 inverted

*/

#include <linux/init.h>
#include <linux/miscdevice.h>

#include <linux/module.h>
#include <linux/kernel.h> /* printk() */

#include <linux/interrupt.h>
#include <linux/gpio.h>

#include <linux/stat.h> /* S_IRUSR S_IRGRP S_IROTH S_IWUSR S_IWGRP S_IWOTH etc. mode flags */

#include <linux/signal.h>  /* Signal numbers etc. */
#include <linux/sched.h>  /* process-ident things. */
#include <linux/sched/signal.h>  /* send_sig_info . */

#include <linux/slab.h>   /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>  /* O_ACCMODE */

#include <linux/uaccess.h>  /* copy_from/to_user */

MODULE_LICENSE("Dual BSD/GPL");

#define FCMBUFSIZE 100

/* These apper in the /sys/kernel/debug/gpio listing  */ 
#define CL_LABEL0 "rtws0"
#define CL_LABEL1 "rtws1"
#define CL_LABEL2 "rtws2"

#define THISFILE "rtws.c"

static int f_open(struct inode *inode, struct file *filp);
static int f_release(struct inode *inode, struct file *filp);
static ssize_t f_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
static void f_exit(void);
static int f_init(void);

/* Structure that declares the access functions. */

struct file_operations rtws_fops = 
{
  read: f_read,
  open: f_open,
  release: f_release
};

/* Misc device structure. */
static struct miscdevice rtws_misc_device 
	= {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "rtws",
	.fops = &rtws_fops,
	.mode = S_IRUSR | S_IRGRP | S_IROTH 
  };
	/* No use for this being write-able when there is no write function! */ 

/* Declaration of the init and exit functions */
module_init(f_init);
module_exit(f_exit);

/* This is essentially read-only so could open as many times as we like */ 

/* Which signal to send and where to send it? */

/* Global variables of the driver. */

static short gpio_r0 = 17;
static short gpio_r1 = 27;
static short gpio_r2 = 22;


/* use:
	gpio_direction_input(gpio_pb);
	value = gpio_get_value(gpio_pb);

	or: 
	gpio_direction_output(gpio_pb, value);
	gpio_set_value(gpio_pb, value);
*/ 

/* Init: Allocate the gpios */
static int f_init(void) 
{
	int rx;

	/* Request the input gpio line and enable the interrupts on them. */
	rx = gpio_request(gpio_r0, CL_LABEL0);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d request failure %d\n", THISFILE, __LINE__, gpio_r0, rx); 
		goto nogpio0;
	}

	rx = gpio_direction_input(gpio_r0);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d input direction request failure %d\n", THISFILE, __LINE__, gpio_r0, rx); 
		goto nogpio0in;
	}

	rx = gpio_request(gpio_r1, CL_LABEL1);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d request failure %d\n", THISFILE, __LINE__, gpio_r1, rx); 
		goto nogpio1;
	}

	rx = gpio_direction_input(gpio_r1);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d input direction request failure %d\n", THISFILE, __LINE__, gpio_r1, rx); 
		goto nogpio1in;
	}

	rx = gpio_request(gpio_r2, CL_LABEL2);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d request failure %d\n", THISFILE, __LINE__, gpio_r2, rx); 
		goto nogpio2;
	}

	rx = gpio_direction_input(gpio_r2);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d input direction request failure %d\n", THISFILE, __LINE__, gpio_r2, rx); 
		goto nogpio2in;
	}

	/* Register the device itself. */
	rx = misc_register(&rtws_misc_device);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): Register misc device failure %d\n", THISFILE, __LINE__, rx); 
		goto nodevice;
	}

	printk(KERN_INFO "%s(%d): Info: misc device minor number = %d\n", 
		THISFILE, __LINE__, rtws_misc_device.minor);

	return 0;

/* Error exits */
nodevice:
nogpio2in:
	gpio_free(gpio_r2); 
nogpio2:
nogpio1in:
	gpio_free(gpio_r1); 
nogpio1:
nogpio0in:
	gpio_free(gpio_r0); 
nogpio0:
	return(rx);
}


static void f_exit(void) 
{
	gpio_free(gpio_r2); 
	gpio_free(gpio_r1); 
	gpio_free(gpio_r0); 

	misc_deregister(&rtws_misc_device);
	printk(KERN_INFO "%s(%d): Removing rtws\n", THISFILE, __LINE__);
}


static int f_open(struct inode *inode, struct file *filp) 
{
	try_module_get(THIS_MODULE); 
	return 0;
}

static int f_release(struct inode *inode, struct file *filp) 
{
	module_put(THIS_MODULE);
	return 0;
}

/* Read the three lines and assemble the return value. */
static ssize_t f_read(struct file *filp, char *buff_out, size_t count, loff_t *f_pos) 
{ 
	size_t wcount;
	char wbuff[FCMBUFSIZE];
	int cuu;

	int ga, gx;
	ga = 0; 

	gx = gpio_get_value(gpio_r0);
	if(gx == 0) ga += 1;

	gx = gpio_get_value(gpio_r1);
	if(gx == 0) ga += 2;

	gx = gpio_get_value(gpio_r2);
	if(gx == 0) ga += 4;

	/* Read: return values of the various statuses:
		Total
		Countdown
		Relay state
		intr setting
	*/

	snprintf(wbuff, sizeof(wbuff), "%u\n", ga); 

	wcount = strlen(wbuff);

	/* Transfering data to user space */ 
	cuu = copy_to_user(buff_out, wbuff, wcount);

	/* Changing reading position as best suits */ 
	if(*f_pos == 0) 
	{ 
		*f_pos += wcount; 
		return wcount; 
	} 
	else 
	{ 
		return 0; 
	}
}



